@@ -80,7 +80,7 @@ BEGIN {
                       initdebug enabledebug enabledebuglevel
                       printdebug debugcmd
                       $printdebug_when_debuglevel $debugcmd_when_debuglevel
-                      $atext_re $dot_atom_text_re $angle_addr_re
+                      $atext_re $dot_atom_text_re $addr_spec_re $angle_addr_re
                       $debugprefix *debuglevel *DEBUG
                       shellquote printcmd messagequote
                       $negate_harmful_gitattrs
@@ -125,8 +125,10 @@ our $debugcmd_when_debuglevel = 1;
 our $atext_re         = qr([[:alnum:]!#$%&'*+\-/=?^_`{|}~])a;
 # This is RFC 5322's 'dot-atom-text' without comments and whitespace.
 our $dot_atom_text_re = qr($atext_re(?:\.|$atext_re)*)a;
+# This is RFC 5322's 'addr-spec' without obsolete syntax.
+our $addr_spec_re = qr($dot_atom_text_re\@$dot_atom_text_re);
 # This is RFC 5322's 'angle-addr' without obsolete syntax.
-our $angle_addr_re = qr(\<($dot_atom_text_re\@$dot_atom_text_re)\>)a;
+our $angle_addr_re = qr(\<($addr_spec_re)\>)a;
 
 our (@git) = qw(git);
 
@@ -235,6 +237,19 @@ sub messagequote ($) {
 }
 
 sub shellquote {
+    # Quote an argument list for use as a fragment of shell text.
+    #
+    # Shell quoting doctrine in dgit.git:
+    #  * perl lists are always unquoted argument lists
+    #  * perl scalars are always individual arguments,
+    #    or if being passed to a shell, quoted shell text.
+    #
+    # So shellquote returns a scalar.
+    #
+    # When invoking ssh-like programs, that concatenate the arguments
+    # with spaces and then treat the result as a shell command, we never
+    # use the concatenation.  We pass the intended script as a single
+    # parameter (which is in accordance with the above doctrine).
     my @out;
     local $_;
     defined or confess __ 'internal error' foreach @_;
@@ -253,7 +268,7 @@ sub shellquote {
 sub printcmd {
     my $fh = shift @_;
     my $intro = shift @_;
-    print $fh $intro." ".(join '', shellquote @_)."\n" or confess "$!";
+    print $fh $intro." ".(shellquote @_)."\n" or confess "$!";
 }
 
 sub debugcmd {
@@ -559,9 +574,9 @@ sub hashfile ($) {
 }
 
 sub git_rev_parse ($;$) {
-    my ($ref, $cmdprefix) = @_;
-    $cmdprefix //= [ ];
-    return cmdoutput @$cmdprefix, qw(git rev-parse), "$ref~0";
+    my ($ref, $cmd_map) = @_;
+    $cmd_map //= sub { @_; };
+    return cmdoutput $cmd_map->(qw(git rev-parse), "$ref~0");
 }
 
 sub changedir_git_toplevel () {
@@ -872,12 +887,13 @@ sub parsecontrol {
 }
 
 sub parsechangelog {
-    # parsechangelog [ \@cmdprefix ], @dpkg_changelog_args
+    # parsechangelog            @dpkg_changelog_args
+    # parsechangelog \&cmd_map, @dpkg_changelog_args
     my $c = Dpkg::Control::Hash->new(name => 'parsed changelog');
     my $p = new IO::Handle;
-    my $prefix = [ ];
-    $prefix = shift @_ if ref $_[0];
-    my @cmd = (@$prefix, qw(dpkg-parsechangelog), @_);
+    my $cmd_map = sub { @_; };
+    $cmd_map = shift @_ if ref $_[0];
+    my @cmd = $cmd_map->(qw(dpkg-parsechangelog), @_);
     open $p, '-|', @cmd or confess "$!";
     $c->parse($p);
     $?=0; $!=0; close $p or failedcmd @cmd;
